program patchelf;

// Copyright Simon Kissel 2004
// This is released as free software under the BSD license.
// So, do whatever you wish with this code.

{$APPTYPE CONSOLE}

uses
  SysUtils, Classes;

type
  Elf32_Addr = Cardinal;
  Elf32_Half = Word;
  Elf32_Off = Cardinal;
  Elf32_Sword = Integer;
  Elf32_Word = Cardinal;

  TElfHeader = packed record
    e_ident: packed array[0..15] of char;
    e_type: Elf32_Half;
    e_machine: Elf32_Half;
    e_version: Elf32_Word;
    e_entry: Elf32_Addr;
    e_phoff: Elf32_Off;
    e_shoff: Elf32_Off;
    e_flags: Elf32_Word;
    e_ehsize: Elf32_Half;
    e_phentsize: Elf32_Half;
    e_phnum: Elf32_Half;
    e_shentsize: Elf32_Half;
    e_shnum: Elf32_Half;
    e_shstrndx: Elf32_Half;
  end;

  TProgramHeader = packed record
    p_type: Elf32_Word;
    p_offset: Elf32_Off;
    p_vaddr: Elf32_Addr;
    p_paddr: Elf32_Addr;
    p_filesz: Elf32_Word;
    p_memsz: Elf32_Word;
    p_flags: Elf32_Word;
    p_align: Elf32_Word;
  end;

var fs:TFileStream;
    ElfHeader:TElfHeader;
    i:integer;
    ProgramHeaders: array of TProgramHeader;
    s:string;
    ProgramHeaderOffset:integer;
    DynamicSectionOffset:integer;
    b:byte;
    filename:string;
    quiet: boolean;
    fixed: boolean;
begin
  if paramcount<1 then
  begin
    writeln('patchelf v1.0 - (c) Simon Kissel 2004');
    writeln('--------------------------------------------------------');
    writeln('This utility will patch Kylix-compiled ELF binaries to');
    writeln('fix a missing executable flag which cause these binaries');
    writeln('to crash on recent Linux kernels (2.6.8+).');
    writeln('');
    writeln('This is free software under the BSD license.');
    writeln('Project homepage at http://crosskylix.untergrund.net');
    writeln('');
    writeln('Usage:');
    writeln('');
    writeln('patchelf [-q] filename');
    writeln('');
    writeln('  -q     Quiet mode (no output)');
    exit;
  end;

  quiet:=false;
  fixed:=false;
  fs:=nil;

  for i:=1 to paramcount do
  begin
    s:=lowercase(paramstr(i));
    if (s='-q') or (s='/q') then quiet:=true else
    begin
      filename:=paramstr(i);
      if not fileexists(filename) then
      begin
        writeln('Error: File '+filename+' does not exist.');
        halt(1);
      end;
      break;
    end;
  end;

  try
    fs:=TFileStream.Create(filename,fmOpenReadWrite or fmShareDenyWrite);
  except
    on e:Exception do
    begin
      writeln('Error: Unable to open '+filename+' for reading and writing.');
      writeln('Check permissions and make sure file is not in use.');
      halt(2);
    end;
  end;

  fs.Read(ElfHeader,sizeof(TElfheader));
  if (ElfHeader.e_ident[1]<>'E') or (ElfHeader.e_ident[2]<>'L') or (ElfHeader.e_ident[3]<>'F') then
  begin
    writeln('Error: This is not a valid linux ELF binary.');
    halt(3);
  end;

  if not quiet then write('Checking ELF Header... ');

  if ElfHeader.e_phentsize=sizeof(TProgramHeader) then
  begin
    if not quiet then writeln('OK');
  end
  else
  begin
    writeln('Error: Unsupported Program Header size.');
    writeln('If this file really is a Kylix compiled ELF executable, please report');
    writeln('back to the author.');
    halt(4);
  end;

  if not quiet then writeln('Reading Program Headers...');

  setlength(ProgramHeaders,ElfHeader.e_phnum);
  ProgramHeaderOffset:=ElfHeader.e_phoff;
  fs.Position:=ProgramHeaderOffset;

  for i:=0 to ElfHeader.e_phnum-1 do
  begin
    fs.Read(ProgramHeaders[i],sizeof(TProgramHeader));
  end;

  if not quiet then writeln('Checking Program Header table...');

  for i:=0 to ElfHeader.e_phnum-1 do
  begin
    if ProgramHeaders[i].p_type=1 then
    begin
      if not quiet then
      begin
        write('  Loadable segment at '+inttohex(ProgramHeaders[i].p_vaddr,8)+'/'+inttohex(ProgramHeaders[i].p_paddr,8)+', Size '+inttohex(ProgramHeaders[i].p_memsz,8)+', Flags ');
        if (ProgramHeaders[i].p_flags AND 4)=4 then write('R');
        if (ProgramHeaders[i].p_flags AND 2)=2 then write('W');
        if (ProgramHeaders[i].p_flags AND 1)=1 then write('E');
        writeln;
      end;
      if (ProgramHeaders[i].p_flags AND 1)<>1 then
      begin
        if not quiet then write('  Missing executable flag, fixing... ');
        DynamicSectionOffset:=ProgramHeaderOffset+sizeof(TProgramHeader)*i+24;
        fs.Position:=DynamicSectionOffset;
        b:=ProgramHeaders[i].p_flags or 1;
        fs.Write(b,1);
        if not quiet then writeln('OK');
        fixed:=true;
      end;
    end;
  end;

  if not quiet then
  begin
    if fixed then
      writeln('File fixed successfully!')
    else
      writeln('There was nothing to fix in this file.');
  end;
  fs.free;
end.
